/*
 * ***** BEGIN GPL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

/** \file blender/editors/interface/resources.c
 *  \ingroup edinterface
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "MEM_guardedalloc.h"

#include "DNA_curve_types.h"
#include "DNA_userdef_types.h"
#include "DNA_screen_types.h"
#include "DNA_space_types.h"
#include "DNA_windowmanager_types.h"

#include "BLI_blenlib.h"
#include "BLI_utildefines.h"
#include "BLI_math.h"

#include "BKE_DerivedMesh.h"
#include "BKE_global.h"
#include "BKE_main.h"
#include "BKE_texture.h"


#include "BIF_gl.h"

#include "UI_interface.h"
#include "UI_interface_icons.h"

#include "interface_intern.h"

/* global for themes */
typedef void (*VectorDrawFunc)(int x, int y, int w, int h, float alpha);

static bTheme *theme_active = NULL;
static int theme_spacetype = SPACE_VIEW3D;
static int theme_regionid = RGN_TYPE_WINDOW;

void ui_resources_init(void)
{
	UI_icons_init(BIFICONID_LAST);
}

void ui_resources_free(void)
{
	UI_icons_free();
}


/* ******************************************************** */
/*    THEMES */
/* ******************************************************** */

const unsigned char *UI_ThemeGetColorPtr(bTheme *btheme, int spacetype, int colorid)
{
	ThemeSpace *ts = NULL;
	static char error[4] = {240, 0, 240, 255};
	static char alert[4] = {240, 60, 60, 255};
	static char headerdesel[4] = {0, 0, 0, 255};
	static char setting = 0;
	const char *cp = error;
	
	if (btheme) {
	
		/* first check for ui buttons theme */
		if (colorid < TH_THEMEUI) {
		
			switch (colorid) {

				case TH_REDALERT:
					cp = alert; break;
			}
		}
		else {
		
			switch (spacetype) {
				case SPACE_BUTS:
					ts = &btheme->tbuts;
					break;
				case SPACE_VIEW3D:
					ts = &btheme->tv3d;
					break;
				case SPACE_IPO:
					ts = &btheme->tipo;
					break;
				case SPACE_FILE:
					ts = &btheme->tfile;
					break;
				case SPACE_NLA:
					ts = &btheme->tnla;
					break;
				case SPACE_ACTION:
					ts = &btheme->tact;
					break;
				case SPACE_SEQ:
					ts = &btheme->tseq;
					break;
				case SPACE_IMAGE:
					ts = &btheme->tima;
					break;
				case SPACE_TEXT:
					ts = &btheme->text;
					break;
				case SPACE_OUTLINER:
					ts = &btheme->toops;
					break;
				case SPACE_INFO:
					ts = &btheme->tinfo;
					break;
				case SPACE_USERPREF:
					ts = &btheme->tuserpref;
					break;
				case SPACE_CONSOLE:
					ts = &btheme->tconsole;
					break;
				case SPACE_TIME:
					ts = &btheme->ttime;
					break;
				case SPACE_NODE:
					ts = &btheme->tnode;
					break;
				case SPACE_LOGIC:
					ts = &btheme->tlogic;
					break;
				case SPACE_CLIP:
					ts = &btheme->tclip;
					break;
				default:
					ts = &btheme->tv3d;
					break;
			}

			switch (colorid) {
				case TH_BACK:
					if (theme_regionid == RGN_TYPE_WINDOW)
						cp = ts->back;
					else if (theme_regionid == RGN_TYPE_CHANNELS)
						cp = ts->list;
					else if (theme_regionid == RGN_TYPE_HEADER)
						cp = ts->header;
					else
						cp = ts->button;
					break;
				case TH_LOW_GRAD:
					cp = ts->gradients.gradient;
					break;
				case TH_HIGH_GRAD:
					cp = ts->gradients.high_gradient;
					break;
				case TH_SHOW_BACK_GRAD:
					cp = &setting;
					setting = ts->gradients.show_grad;
					break;
				case TH_TEXT:
					if (theme_regionid == RGN_TYPE_WINDOW)
						cp = ts->text;
					else if (theme_regionid == RGN_TYPE_CHANNELS)
						cp = ts->list_text;
					else if (theme_regionid == RGN_TYPE_HEADER)
						cp = ts->header_text;
					else
						cp = ts->button_text;
					break;
				case TH_TEXT_HI:
					if (theme_regionid == RGN_TYPE_WINDOW)
						cp = ts->text_hi;
					else if (theme_regionid == RGN_TYPE_CHANNELS)
						cp = ts->list_text_hi;
					else if (theme_regionid == RGN_TYPE_HEADER)
						cp = ts->header_text_hi;
					else
						cp = ts->button_text_hi;
					break;
				case TH_TITLE:
					if (theme_regionid == RGN_TYPE_WINDOW)
						cp = ts->title;
					else if (theme_regionid == RGN_TYPE_CHANNELS)
						cp = ts->list_title;
					else if (theme_regionid == RGN_TYPE_HEADER)
						cp = ts->header_title;
					else
						cp = ts->button_title;
					break;

				case TH_HEADER:
					cp = ts->header; break;
				case TH_HEADERDESEL:
					/* we calculate a dynamic builtin header deselect color, also for pulldowns... */
					cp = ts->header;
					headerdesel[0] = cp[0] > 10 ? cp[0] - 10 : 0;
					headerdesel[1] = cp[1] > 10 ? cp[1] - 10 : 0;
					headerdesel[2] = cp[2] > 10 ? cp[2] - 10 : 0;
					cp = headerdesel;
					break;
				case TH_HEADER_TEXT:
					cp = ts->header_text; break;
				case TH_HEADER_TEXT_HI:
					cp = ts->header_text_hi; break;

				case TH_PANEL_HEADER:
					cp = ts->panelcolors.header; break;
				case TH_PANEL_BACK:
					cp = ts->panelcolors.back; break;
				case TH_PANEL_SHOW_HEADER:
					cp = &setting;
					setting = ts->panelcolors.show_header;
					break;
				case TH_PANEL_SHOW_BACK:
					cp = &setting;
					setting = ts->panelcolors.show_back;
					break;
					
				case TH_BUTBACK:
					cp = ts->button; break;
				case TH_BUTBACK_TEXT:
					cp = ts->button_text; break;
				case TH_BUTBACK_TEXT_HI:
					cp = ts->button_text_hi; break;

				case TH_SHADE1:
					cp = ts->shade1; break;
				case TH_SHADE2:
					cp = ts->shade2; break;
				case TH_HILITE:
					cp = ts->hilite; break;
				
				case TH_GRID:
					cp = ts->grid; break;
				case TH_WIRE:
					cp = ts->wire; break;
				case TH_WIRE_EDIT:
					cp = ts->wire_edit; break;
				case TH_LAMP:
					cp = ts->lamp; break;
				case TH_SPEAKER:
					cp = ts->speaker; break;
				case TH_CAMERA:
					cp = ts->camera; break;
				case TH_EMPTY:
					cp = ts->empty; break;
				case TH_SELECT:
					cp = ts->select; break;
				case TH_ACTIVE:
					cp = ts->active; break;
				case TH_GROUP:
					cp = ts->group; break;
				case TH_GROUP_ACTIVE:
					cp = ts->group_active; break;
				case TH_TRANSFORM:
					cp = ts->transform; break;
				case TH_VERTEX:
					cp = ts->vertex; break;
				case TH_VERTEX_SELECT:
					cp = ts->vertex_select; break;
				case TH_VERTEX_UNREFERENCED:
					cp = ts->vertex_unreferenced; break;
				case TH_VERTEX_SIZE:
					cp = &ts->vertex_size; break;
				case TH_OUTLINE_WIDTH:
					cp = &ts->outline_width; break;
				case TH_EDGE:
					cp = ts->edge; break;
				case TH_EDGE_SELECT:
					cp = ts->edge_select; break;
				case TH_EDGE_SEAM:
					cp = ts->edge_seam; break;
				case TH_EDGE_SHARP:
					cp = ts->edge_sharp; break;
				case TH_EDGE_CREASE:
					cp = ts->edge_crease; break;
				case TH_EDITMESH_ACTIVE:
					cp = ts->editmesh_active; break;
				case TH_EDGE_FACESEL:
					cp = ts->edge_facesel; break;
				case TH_FACE:
					cp = ts->face; break;
				case TH_FACE_SELECT:
					cp = ts->face_select; break;
				case TH_FACE_DOT:
					cp = ts->face_dot; break;
				case TH_FACEDOT_SIZE:
					cp = &ts->facedot_size; break;
				case TH_DRAWEXTRA_EDGELEN:
					cp = ts->extra_edge_len; break;
				case TH_DRAWEXTRA_EDGEANG:
					cp = ts->extra_edge_angle; break;
				case TH_DRAWEXTRA_FACEAREA:
					cp = ts->extra_face_area; break;
				case TH_DRAWEXTRA_FACEANG:
					cp = ts->extra_face_angle; break;
				case TH_NORMAL:
					cp = ts->normal; break;
				case TH_VNORMAL:
					cp = ts->vertex_normal; break;
				case TH_BONE_SOLID:
					cp = ts->bone_solid; break;
				case TH_BONE_POSE:
					cp = ts->bone_pose; break;
				case TH_BONE_POSE_ACTIVE:
					cp = ts->bone_pose_active; break;
				case TH_STRIP:
					cp = ts->strip; break;
				case TH_STRIP_SELECT:
					cp = ts->strip_select; break;
				case TH_CFRAME:
					cp = ts->cframe; break;
				case TH_NURB_ULINE:
					cp = ts->nurb_uline; break;
				case TH_NURB_VLINE:
					cp = ts->nurb_vline; break;
				case TH_NURB_SEL_ULINE:
					cp = ts->nurb_sel_uline; break;
				case TH_NURB_SEL_VLINE:
					cp = ts->nurb_sel_vline; break;
				case TH_ACTIVE_SPLINE:
					cp = ts->act_spline; break;
				case TH_LASTSEL_POINT:
					cp = ts->lastsel_point; break;
				case TH_HANDLE_FREE:
					cp = ts->handle_free; break;
				case TH_HANDLE_AUTO:
					cp = ts->handle_auto; break;
				case TH_HANDLE_AUTOCLAMP:
					cp = ts->handle_auto_clamped; break;
				case TH_HANDLE_VECT:
					cp = ts->handle_vect; break;
				case TH_HANDLE_ALIGN:
					cp = ts->handle_align; break;
				cp= ts->syntaxv; break;
			case TH_NODE_GROUP:
				cp= ts->syntaxc; break;
			case TH_NODE_CURVING:
				cp= &ts->noodle_curving; break;

