/*
 * ***** BEGIN GPL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The Original Code is Copyright (C) 2007 Blender Foundation.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): Lukas Toenne.
 *
 * ***** END GPL LICENSE BLOCK *****
 */

/** \file blender/nodes/intern/node_common.c
 *  \ingroup nodes
 */


#include <string.h>

#include "DNA_node_types.h"

#include "BLI_listbase.h"
#include "BLI_string.h"
#include "BLI_utildefines.h"

#include "BKE_global.h"
#include "BKE_library.h"
#include "BKE_main.h"
#include "BLI_math.h"
#include "BKE_node.h"
#include "BKE_utildefines.h"

#include "RNA_access.h"
#include "RNA_types.h"

#include "MEM_guardedalloc.h"

#include "node_common.h"
#include "node_exec.h"
#include "NOD_socket.h"

/**** Group ****/

bNodeSocket *node_group_find_input_socket(bNode *groupnode, int own_index)
{
	bNodeSocket *sock;
	for (sock=groupnode->inputs.first; sock; sock=sock->next)
		if (sock->own_index == own_index)
			return sock;
	return NULL;
}

bNodeSocket *node_group_find_output_socket(bNode *groupnode, int own_index)
{
	bNodeSocket *sock;
	for (sock=groupnode->outputs.first; sock; sock=sock->next)
		if (sock->own_index == own_index)
			return sock;
	return NULL;
}

bNodeSocket *node_group_add_extern_socket(bNodeTree *UNUSED(ntree), ListBase *lb, int in_out, struct bNodeSocket *gsock)
{
	bNodeSocket *sock;
	
	if (gsock->flag & SOCK_INTERNAL)
		return NULL;
	
	sock= MEM_callocN(sizeof(bNodeSocket), "sock");
	
	/* make a copy of the group socket */
	*sock = *gsock;
	sock->link = NULL;
	sock->next = sock->prev = NULL;
	sock->new_sock = NULL;
	
	/* group sockets are dynamically added */
	sock->flag |= SOCK_DYNAMIC;
	
	sock->own_index = gsock->own_index;
	sock->limit = (in_out==SOCK_IN ? 1 : 0xFFF);
	
	sock->default_value = node_socket_make_default_value(sock->type);
	node_socket_copy_default_value(sock->type, sock->default_value, gsock->default_value);
	
	if(lb)
		BLI_addtail(lb, sock);
	
	return sock;
}

/* groups display their internal tree name as label */
const char *node_group_label(bNode *node)
{
	return (node->id)? node->id->name+2: "Missing Datablock";
}

int node_group_valid(bNodeTree *ntree, bNodeTemplate *ntemp)
{
	bNodeTemplate childtemp;
	bNode *node;
	
	/* regular groups cannot be recursive */
	if (ntree == ntemp->ngroup)
		return 0;
	
	/* make sure all children are valid */
	for (node=ntemp->ngroup->nodes.first; node; node=node->next) {
		childtemp = nodeMakeTemplate(node);
		if (!nodeValid(ntree, &childtemp))
			return 0;
	}
	
	return 1;
}

bNodeTemplate node_group_template(bNode *node)
{
	bNodeTemplate ntemp;
	ntemp.type = NODE_GROUP;
	ntemp.ngroup = (bNodeTree*)node->id;
	return ntemp;
}

void node_group_init(bNodeTree *ntree, bNode *node, bNodeTemplate *ntemp)
{
	node->id = (ID*)ntemp->ngroup;
	
	node_group_verify(ntree, node, (ID*)ntemp->ngroup);
}

static bNodeSocket *group_verify_socket(bNodeTree *ntree, ListBase *lb, int in_out, bNodeSocket *gsock)
{
	bNodeSocket *sock;
	
	/* group sockets tagged as internal are not exposed ever */
	if (gsock->flag & SOCK_INTERNAL)
		return NULL;
	
	for(sock= lb->first; sock; sock= sock->next) {
		if(sock->own_index==gsock->own_index)
				break;
	}
	if(sock) {
		BLI_strncpy(sock->name, gsock->name, sizeof(sock->name));
		
		if (gsock->type != sock->type)
			nodeSocketSetType(sock, gsock->type);
		
		/* update type details from internal socket */
		node_socket_verify_default_value_type(sock->type, sock->default_value, gsock->default_value);
		
		/* XXX hack: group socket input/output roles are inverted internally,
		 * need to change the limit value when making actual node sockets from them.
		 */
		sock->limit = (in_out==SOCK_IN ? 1 : 0xFFF);
		
		BLI_remlink(lb, sock);
		
		return sock;
	}
	else {
		return node_group_add_extern_socket(ntree, NULL, in_out, gsock);
	}
}

static void group_verify_socket_list(bNodeTree *ntree, bNode *gnode, ListBase *lb, int in_out, bNodeTree *ngroup)
{
	bNodeProxyItem *item;
	bNodeSocket *sock, *nextsock;
	
	/* step by step compare */
	
	item = (in_out==SOCK_IN ? ngroup->proxy_inputs.first : ngroup->proxy_outputs.first);
	for (; item; item=item->next) {
		bNodeSocket *gsock = (in_out==SOCK_IN ? item->node->outputs.first : item->node->inputs.first);
		/* abusing new_sock pointer for verification here! only used inside this function */
		gsock->new_sock = group_verify_socket(ntree, lb, in_out, gsock);
	}
	/* leftovers are removed */
	for (sock=lb->first; sock; sock=nextsock) {
		nextsock=sock->next;
		if (sock->flag & SOCK_DYNAMIC)
			nodeRemoveSocket(ntree, gnode, sock);
	}
	/* and we put back the verified sockets */
	item = (in_out==SOCK_IN ? ngroup->proxy_inputs.first : ngroup->proxy_outputs.first);
	for (; item; item=item->next) {
		bNodeSocket *gsock = (in_out==SOCK_IN ? item->node->outputs.first : item->node->inputs.first);
		if (gsock->new_sock) {
			BLI_addtail(lb, gsock->new_sock);
			gsock->new_sock = NULL;
		}
	}
}

/* make sure all group node in ntree, which use ngroup, are sync'd */
void node_group_verify(struct bNodeTree *ntree, struct bNode *node, struct ID *id)
{
	/* check inputs and outputs, and remove or insert them */
	if (node->id==id) {
		bNodeTree *ngroup= (bNodeTree*)node->id;
		group_verify_socket_list(ntree, node, &node->inputs, SOCK_IN, ngroup);
		group_verify_socket_list(ntree, node, &node->outputs, SOCK_OUT, ngroup);
	}
}

struct bNodeTree *node_group_edit_get(bNode *node)
{
	if (node->flag & NODE_GROUP_EDIT)
		return (bNodeTree*)node->id;
	else
		return NULL;
}

struct bNodeTree *node_group_edit_set(bNode *node, int edit)
{
	if (edit) {
		bNodeTree *ngroup= (bNodeTree*)node->id;
		if (ngroup) {
			if(ngroup->id.lib)
				ntreeMakeLocal(ngroup);
			
			node->flag |= NODE_GROUP_EDIT;
		}
		return ngroup;
	}
	else {
		node->flag &= ~NODE_GROUP_EDIT;
		return NULL;
	}
}

void node_group_edit_clear(bNode *node)
{
	bNodeTree *ngroup= (bNodeTree*)node->id;
	bNode *inode;
	
	node->flag &= ~NODE_GROUP_EDIT;
	
	if (ngroup)
		for (inode=ngroup->nodes.first; inode; inode=inode->next)
			nodeGroupEditClear(inode);
}

#if 0
/**** For Loop ****/

/* Essentially a group node with slightly different behavior.
 * The internal tree is executed several times, with each output being re-used
 * as an input in the next iteration. For this purpose, input and output socket
 * lists are kept identical!
 */

bNodeTemplate node_forloop_template(bNode *node)
{
	bNodeTemplate ntemp;
	ntemp.type = NODE_FORLOOP;
	ntemp.ngroup = (bNodeTree*)node->id;
	return ntemp;
}

void node_forloop_init(bNodeTree *ntree, bNode *node, bNodeTemplate *ntemp)
{
	bNodeSocket *sock;
	
	node->id = (ID*)ntemp->ngroup;
	
	sock = nodeAddSocket(ntree, node, SOCK_IN, "Iterations", SOCK_FLOAT);
	node_socket_set_default_value_float(sock->default_value, PROP_UNSIGNED, 1, 0, 10000);
	
	/* NB: group socket input/output roles are inverted internally!
	 * Group "inputs" work as outputs in links and vice versa.
	 */
	if (ntemp->ngroup) {
		bNodeSocket *gsock;
		for (gsock=ntemp->ngroup->inputs.first; gsock; gsock=gsock->next)
			node_group_add_extern_socket(ntree, &node->inputs, SOCK_IN, gsock);
		for (gsock=ntemp->ngroup->outputs.first; gsock; gsock=gsock->next)
			node_group_add_extern_socket(ntree, &node->outputs, SOCK_OUT, gsock);
	}
}

void node_forloop_init_tree(bNodeTree *ntree)
{
	bNodeSocket *sock;
	sock = node_group_add_socket(ntree, "Iteration", SOCK_FLOAT, SOCK_IN);
	sock->flag |= SOCK_INTERNAL;
}

static void loop_sync(bNodeTree *ntree, int sync_in_out)
{
	bNodeSocket *sock, *sync, *nsync, *mirror;
	ListBase *sync_lb;
	
	if (sync_in_out==SOCK_IN) {
		sock = ntree->outputs.first;
		
		sync = ntree->inputs.first;
		sync_lb = &ntree->inputs;
	}
	else {
		sock = ntree->inputs.first;
		
		sync = ntree->outputs.first;
		sync_lb = &ntree->outputs;
	}
	
	/* NB: the sock->storage pointer is used here directly to store the own_index int
	 * out the mirrored socket counterpart!
	 */
	
	while (sock) {
		/* skip static and internal sockets on the sync side (preserves socket order!) */
		while (sync && ((sync->flag & SOCK_INTERNAL) || !(sync->flag & SOCK_DYNAMIC)))
			sync = sync->next;
		
		if (sync && !(sync->flag & SOCK_INTERNAL) && (sync->flag & SOCK_DYNAMIC)) {
			if (sock->storage==NULL) {
				/* if mirror index is 0, the sockets is newly added and a new mirror must be created. */
				mirror = node_group_expose_socket(ntree, sock, sync_in_out);
				/* store the mirror index */
				sock->storage = SET_INT_IN_POINTER(mirror->own_index);
				mirror->storage = SET_INT_IN_POINTER(sock->own_index);
				/* move mirror to the right place */
				BLI_remlink(sync_lb, mirror);
				if (sync)
					BLI_insertlinkbefore(sync_lb, sync, mirror);
				else
					BLI_addtail(sync_lb, mirror);
			}
			else {
				/* look up the mirror socket */
				for (mirror=sync; mirror; mirror=mirror->next)
					if (mirror->own_index == GET_INT_FROM_POINTER(sock->storage))
						break;
				/* make sure the name is the same (only for identification by user, no deeper meaning) */
				BLI_strncpy(mirror->name, sock->name, sizeof(mirror->name));
				/* fix the socket order if necessary */
				if (mirror != sync) {
					BLI_remlink(sync_lb, mirror);
					BLI_insertlinkbefore(sync_lb, sync, mirror);
				}
				else
					sync = sync->next;
			}
		}
		
		sock = sock->next;
	}
	
	/* remaining sockets in sync_lb are leftovers from deleted sockets, remove them */
	while (sync) {
		nsync = sync->next;
		if (!(sync->flag & SOCK_INTERNAL) && (sync->flag & SOCK_DYNAMIC))
			node_group_remove_socket(ntree, sync, sync_in_out);
		sync = nsync;
	}
}

void node_loop_update_tree(bNodeTree *ngroup)
{
	/* make sure inputs & outputs are identical */
	if (ngroup->update & NTREE_UPDATE_GROUP_IN)
		loop_sync(ngroup, SOCK_OUT);
	if (ngroup->update & NTREE_UPDATE_GROUP_OUT)
		loop_sync(ngroup, SOCK_IN);
}

void node_whileloop_init(bNodeTree *ntree, bNode *node, bNodeTemplate *ntemp)
{
	bNodeSocket *sock;
	
	node->id = (ID*)ntemp->ngroup;
	
	sock = nodeAddSocket(ntree, node, SOCK_IN, "Condition", SOCK_FLOAT);
	node_socket_set_default_value_float(sock->default_value, PROP_NONE, 1, 0, 1);
	
	/* max iterations */
	node->custom1 = 10000;
	
	/* NB: group socket input/output roles are inverted internally!
	 * Group "inputs" work as outputs in links and vice versa.
	 */
	if (ntemp->ngroup) {
		bNodeSocket *gsock;
		for (gsock=ntemp->ngroup->inputs.first; gsock; gsock=gsock->next)
			node_group_add_extern_socket(ntree, &node->inputs, SOCK_IN, gsock);
		for (gsock=ntemp->ngroup->outputs.first; gsock; gsock=gsock->next)
			node_group_add_extern_socket(ntree, &node->outputs, SOCK_OUT, gsock);
	}
}

void node_whileloop_init_tree(bNodeTree *ntree)
{
	bNodeSocket *sock;
	sock = node_group_add_socket(ntree, "Condition", SOCK_FLOAT, SOCK_OUT);
	sock->flag |= SOCK_INTERNAL;
}

bNodeTemplate node_whileloop_template(bNode *node)
{
	bNodeTemplate ntemp;
	ntemp.type = NODE_WHILELOOP;
	ntemp.ngroup = (bNodeTree*)node->id;
	return ntemp;
}
#endif

/**** FRAME ****/

void register_node_type_frame(bNodeTreeType *ttype)
{
	/* frame type is used for all tree types, needs dynamic allocation */
	bNodeType *ntype= MEM_callocN(sizeof(bNodeType), "frame node type");

	node_type_base(ttype, ntype, NODE_FRAME, "Frame", NODE_CLASS_LAYOUT, NODE_BACKGROUND);
	node_type_size(ntype, 150, 100, 0);
	node_type_compatibility(ntype, NODE_OLD_SHADING|NODE_NEW_SHADING);
	
	ntype->needs_free = 1;
	nodeRegisterType(ttype, ntype);
}

/**** PROXY_INPUT / PROXY_OUTPUT ****/

static void node_proxy_input_init(bNodeTree *ntree, bNode *node, bNodeTemplate *UNUSED(ntemp))
{
	bNodeSocketType *stype;
	bNodeSocket *sock;
	
	/* node tree input becomes output socket on the node */
	stype = ntreeGetSocketType(SOCK_FLOAT);
	sock = nodeAddSocket(ntree, node, SOCK_OUT, stype->ui_name, SOCK_FLOAT);
	
	/* assign new unique index */
	sock->own_index = ntree->cur_index++;
}

static void node_proxy_input_copy(bNodeTree *dest_ntree, bNode *dest_node, bNode *UNUSED(src_node))
{
	/* socket has already been copied */
	bNodeSocket *sock = (bNodeSocket*)dest_node->outputs.first;
	
	/* assign new unique index */
	sock->own_index = dest_ntree->cur_index++;
}

static const char *node_proxy_input_label(bNode *node)
{
	bNodeSocket *sock = node->outputs.first;
	return sock->name;
}

void register_node_type_proxy_input(struct bNodeTreeType *ttype)
{
	/* used for all tree types, needs dynamic allocation */
	bNodeType *ntype= MEM_callocN(sizeof(bNodeType), "node type");
	
	node_type_base(ttype, ntype, NODE_PROXY_INPUT, "Proxy Input", NODE_CLASS_INPUT, NODE_OPTIONS);
	node_type_size(ntype, 140, 80, 200);
	node_type_init(ntype, node_proxy_input_init);
	node_type_storage(ntype, NULL, NULL, node_proxy_input_copy);
	node_type_compatibility(ntype, NODE_OLD_SHADING|NODE_NEW_SHADING);
	node_type_label(ntype, node_proxy_input_label);
	
	ntype->needs_free = 1;
	nodeRegisterType(ttype, ntype);
}

static void node_proxy_output_init(bNodeTree *ntree, bNode *node, bNodeTemplate *UNUSED(ntemp))
{
	bNodeSocketType *stype;
	bNodeSocket *sock;
	
	/* node tree output becomes input socket on the node */
	stype = ntreeGetSocketType(SOCK_FLOAT);
	sock = nodeAddSocket(ntree, node, SOCK_IN, stype->ui_name, SOCK_FLOAT);
	
	/* assign new unique index */
	sock->own_index = ntree->cur_index++;
	
	/* output proxy nodes automatically update type by default */
	node->custom1 |= NODE_PROXY_AUTOTYPE;
}

static void node_proxy_output_copy(bNodeTree *dest_ntree, bNode *dest_node, bNode *UNUSED(src_node))
{
	/* socket has already been copied */
	bNodeSocket *sock = (bNodeSocket*)dest_node->inputs.first;
	
	/* assign new unique index */
	sock->own_index = dest_ntree->cur_index++;
}

static void node_proxy_output_update(bNodeTree *ntree, bNode *node)
{
	if (node->custom1 & NODE_PROXY_AUTOTYPE) {
		/* automatically update the socket type based on linked input */
		bNodeSocket *sock = (bNodeSocket*)node->inputs.first;
		if (sock->link && sock->link->fromsock->type != sock->type) {
			nodeSocketSetType(sock, sock->link->fromsock->type);
			
			ntree->update |= NTREE_UPDATE_GROUP_OUT;
		}
	}
}

static const char *node_proxy_output_label(bNode *node)
{
	bNodeSocket *sock = node->inputs.first;
	return sock->name;
}

void register_node_type_proxy_output(struct bNodeTreeType *ttype)
{
	/* used for all tree types, needs dynamic allocation */
	bNodeType *ntype= MEM_callocN(sizeof(bNodeType), "node type");
	
	node_type_base(ttype, ntype, NODE_PROXY_OUTPUT, "Proxy Output", NODE_CLASS_OUTPUT, NODE_OPTIONS);
	node_type_size(ntype, 140, 80, 200);
	node_type_init(ntype, node_proxy_output_init);
	node_type_storage(ntype, NULL, NULL, node_proxy_output_copy);
	node_type_compatibility(ntype, NODE_OLD_SHADING|NODE_NEW_SHADING);
	node_type_label(ntype, node_proxy_output_label);
	node_type_update(ntype, node_proxy_output_update, NULL);
	
	ntype->needs_free = 1;
	nodeRegisterType(ttype, ntype);
}
