/*
 * ***** BEGIN GPL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The Original Code is Copyright (C) 2007 Blender Foundation.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): Lukas Toennne
 *
 * ***** END GPL LICENSE BLOCK *****
 */

/** \file blender/nodes/intern/node_socket.c
 *  \ingroup nodes
 */

#include <limits.h>

#include "DNA_node_types.h"

#include "BLI_listbase.h"
#include "BLI_math.h"
#include "BLI_string.h"
#include "BLI_utildefines.h"
#include "BLI_string.h"

#include "BKE_node.h"
#include "BKE_idprop.h"

#include "RNA_access.h"
#include "RNA_define.h"
#include "RNA_types.h"

#include "MEM_guardedalloc.h"

#include "NOD_socket.h"

struct Main;

struct bNodeSocket *node_add_socket_from_template(struct bNodeTree *ntree, struct bNode *node, struct bNodeSocketTemplate *stemp, int in_out)
{
	bNodeSocket *sock = nodeAddStaticSocket(ntree, node, in_out, stemp->type, stemp->subtype, stemp->identifier, stemp->name);
	
	sock->flag |= stemp->flag;
	
	/* initialize default_value */
	switch (stemp->type) {
		case SOCK_FLOAT:
		{
			bNodeSocketValueFloat *dval = sock->default_value;
			dval->value = stemp->val1;
			dval->min = stemp->min;
			dval->max = stemp->max;
			break;
		}
		case SOCK_INT:
		{
			bNodeSocketValueInt *dval = sock->default_value;
			dval->value = (int)stemp->val1;
			dval->min = (int)stemp->min;
			dval->max = (int)stemp->max;
			break;
		}
		case SOCK_BOOLEAN:
		{
			bNodeSocketValueBoolean *dval = sock->default_value;
			dval->value = (int)stemp->val1;
			break;
		}
		case SOCK_VECTOR:
		{
			bNodeSocketValueVector *dval = sock->default_value;
			dval->value[0] = stemp->val1;
			dval->value[1] = stemp->val2;
			dval->value[2] = stemp->val3;
			dval->min = stemp->min;
			dval->max = stemp->max;
			break;
		}
		case SOCK_RGBA:
		{
			bNodeSocketValueRGBA *dval = sock->default_value;
			dval->value[0] = stemp->val1;
			dval->value[1] = stemp->val2;
			dval->value[2] = stemp->val3;
			dval->value[3] = stemp->val4;
			break;
		}
	}
	
	return sock;
}

static bNodeSocket *verify_socket_template(bNodeTree *ntree, bNode *node, int in_out, ListBase *socklist, bNodeSocketTemplate *stemp)
{
	bNodeSocket *sock;
	
	for (sock = socklist->first; sock; sock = sock->next) {
		if (strncmp(sock->name, stemp->name, NODE_MAXSTR) == 0)
			break;
	}
	if (sock) {
		sock->type = stemp->type;
		sock->limit = (stemp->limit == 0 ? 0xFFF : stemp->limit);
		sock->flag |= stemp->flag;
		
		BLI_remlink(socklist, sock);
		
		return sock;
	}
	else {
		/* no socket for this template found, make a new one */
		sock = node_add_socket_from_template(ntree, node, stemp, in_out);
		/* remove the new socket from the node socket list first,
		 * will be added back after verification.
		 */
		BLI_remlink(socklist, sock);
	}
	
	return sock;
}

static void verify_socket_template_list(bNodeTree *ntree, bNode *node, int in_out, ListBase *socklist, bNodeSocketTemplate *stemp_first)
{
	bNodeSocket *sock, *nextsock;
	bNodeSocketTemplate *stemp;
	
	/* no inputs anymore? */
	if (stemp_first == NULL) {
		for (sock = (bNodeSocket *)socklist->first; sock; sock = nextsock) {
			nextsock = sock->next;
			nodeRemoveSocket(ntree, node, sock);
		}
	}
	else {
		/* step by step compare */
		stemp = stemp_first;
		while (stemp->type != -1) {
			stemp->sock = verify_socket_template(ntree, node, in_out, socklist, stemp);
			stemp++;
		}
		/* leftovers are removed */
		for (sock = (bNodeSocket *)socklist->first; sock; sock = nextsock) {
			nextsock = sock->next;
			nodeRemoveSocket(ntree, node, sock);
		}
		
		/* and we put back the verified sockets */
		stemp = stemp_first;
		if (socklist->first) {
			/* some dynamic sockets left, store the list start
			 * so we can add static sockets infront of it.
			 */
			sock = socklist->first;
			while (stemp->type != -1) {
				/* put static sockets infront of dynamic */
				BLI_insertlinkbefore(socklist, sock, stemp->sock);
				stemp++;
			}
		}
		else {
			while (stemp->type != -1) {
				BLI_addtail(socklist, stemp->sock);
				stemp++;
			}
		}
	}
}

void node_verify_socket_templates(bNodeTree *ntree, bNode *node)
{
	bNodeType *ntype = node->typeinfo;
	/* Don't try to match socket lists when there are no templates.
	 * This prevents group node sockets from being removed, without the need to explicitly
	 * check the node type here.
	 */
	if (ntype) {
		if (ntype->inputs && ntype->inputs[0].type >= 0)
			verify_socket_template_list(ntree, node, SOCK_IN, &node->inputs, ntype->inputs);
		if (ntype->outputs && ntype->outputs[0].type >= 0)
			verify_socket_template_list(ntree, node, SOCK_OUT, &node->outputs, ntype->outputs);
	}
}


void node_socket_init_default_value(bNodeSocket *sock)
{
	int type = sock->typeinfo->type;
	int subtype = sock->typeinfo->subtype;
	
	if (sock->default_value)
		return; /* already initialized */
	
	switch (type) {
		case SOCK_FLOAT:
		{
			bNodeSocketValueFloat *dval = MEM_callocN(sizeof(bNodeSocketValueFloat), "node socket value float");
			dval->subtype = subtype;
			dval->value = 0.0f;
			dval->min = -FLT_MAX;
			dval->max = FLT_MAX;
		
			sock->default_value = dval;
			break;
		}
		case SOCK_INT:
		{
			bNodeSocketValueInt *dval = MEM_callocN(sizeof(bNodeSocketValueInt), "node socket value int");
			dval->subtype = subtype;
			dval->value = 0;
			dval->min = INT_MIN;
			dval->max = INT_MAX;
		
			sock->default_value = dval;
			break;
		}
		case SOCK_BOOLEAN:
		{
			bNodeSocketValueBoolean *dval = MEM_callocN(sizeof(bNodeSocketValueBoolean), "node socket value bool");
			dval->value = false;
		
			sock->default_value = dval;
			break;
		}
		case SOCK_VECTOR:
		{
			static float default_value[] = { 0.0f, 0.0f, 0.0f };
			bNodeSocketValueVector *dval = MEM_callocN(sizeof(bNodeSocketValueVector), "node socket value vector");
			dval->subtype = subtype;
			copy_v3_v3(dval->value, default_value);
			dval->min = -FLT_MAX;
			dval->max = FLT_MAX;
		
			sock->default_value = dval;
			break;
		}
		case SOCK_RGBA:
		{
			static float default_value[] = { 0.0f, 0.0f, 0.0f, 1.0f };
			bNodeSocketValueRGBA *dval = MEM_callocN(sizeof(bNodeSocketValueRGBA), "node socket value color");
			copy_v4_v4(dval->value, default_value);
		
			sock->default_value = dval;
			break;
		}
		case SOCK_STRING:
		{
			bNodeSocketValueString *dval = MEM_callocN(sizeof(bNodeSocketValueString), "node socket value string");
			dval->subtype = subtype;
			dval->value[0] = '\0';
		
			sock->default_value = dval;
			break;
		}
	}
}

void node_socket_copy_default_value(bNodeSocket *to, bNodeSocket *from)
}


void node_socket_copy_default_value(int type, void *to_default_value, void *from_default_value)
{
	/* XXX only one of these pointers is valid! just putting them here for convenience */
